/*
 * bdremoteircd: a lirc daemon for the Sony BD Remote
 * Copyright (C) 2010 D.Herrendoerfer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/hidraw.h>
#include <linux/input.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include "remote.h"

char *version = "bdremotelircd-0.5";

int sockfd, rfd, FB_present;
char *lircdfile = "/dev/lircd";
char *progname = "lirc.c";
mode_t permission=S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH;
int clients[16] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
socklen_t clilen;
struct sockaddr client_addr;
int num_clients = 0;
int  ipid = 0;
int  debug = 0;
int  nofork = 0;
char lirc_message[128];
int key_down=0;

inline int max(int a,int b) 
{
	return(a>b ? a:b);
}

void nolinger(int sock)
{
        static struct linger  linger = {0, 0};
        int lsize  = sizeof(struct linger);
        setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *)&linger, lsize);
}

int setupSocket()
{
        int ret;
        int new = 0;
        struct sockaddr_un serv_addr;
        struct stat s;

        /* create socket */
        sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
        if (sockfd == -1) {
		fprintf(stderr, "%s: could not create socket\n", progname);
		perror(progname);
		goto start_server_failed0;
	}

        /*
         *         get owner, permissions, etc.
         *         so new socket can be the same since we
         *         have to delete the old socket.
         **/
	ret = stat(lircdfile, &s);
	if (ret == -1 && errno != ENOENT) {
		fprintf(stderr,
			"%s: could not get file information for %s\n",
                        progname, lircdfile);
		perror(progname);
		goto start_server_failed1;
	}
	if (ret != -1) {
		ret = unlink(lircdfile);
		if (ret == -1) {
			fprintf(stderr, "%s: could not delete %s\n",
				progname, lircdfile);
				perror(NULL);
				goto start_server_failed1;
		}
	}

	serv_addr.sun_family = AF_UNIX;
	strcpy(serv_addr.sun_path, lircdfile);
	if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr))
		== -1) {
		fprintf(stderr, "%s: could not assign address to socket\n",
			progname);
		perror(progname);
		goto start_server_failed1;
	}

	if (new ?
	    chmod(lircdfile, permission) :
	    (chmod(lircdfile, s.st_mode) == -1 ||
	    chown(lircdfile, s.st_uid, s.st_gid) == -1)
	    ) {
		fprintf(stderr, "%s: could not set file permissions\n",
		progname);
		perror(progname);
		goto start_server_failed1;
	}

	listen(sockfd, 3);
	nolinger(sockfd);

	return 0;

start_server_failed0:
start_server_failed1:
	return -1;
}

static void help(void)
{
        printf("bdremotelircd - Sony BD Remote lirc daemon.\n");
        printf("Usage:\n\tbdremotelircd [options]\n\n");
        printf("Options:\n"
                "\v-v                   Print version.\n"
                "\t-n                   Dont fork to background.\n"
                "\t-d                   Print lots of debugging information.\n"
                "\t-h, 	                Help message.\n\n");
}

int scanforfb(char *usename)
{
        int fd, i;
	char device[256] = "";
	char name[256] = "Unknown";
	if ( debug )
		printf ("Scanning for PS3 BD remote:\n");

	for (i=0 ; i<16 ; i++) {
		sprintf(device,"/dev/hidraw%i",i);

		if ((fd = open (device, O_RDONLY)) == -1) {
			if ( debug )
			printf("  can't open %s - assuming as unused.\n",device);
		}
		else {
			ioctl (fd, HIDIOCGRAWNAME (sizeof (name)), name);
			if ( debug )
				printf("  hidraw device /dev/hidraw%i (%s).\n",i,name);
                        
			if (!strcmp (name,usename)) {
				if ( debug )
					printf ("    !! found PS3 BD remote !!\n");

				return fd;
			}
			close(fd);
		}
	}
	return 0;
}

void add_client()
{
	int nfd;
	nfd=accept(sockfd,(struct sockaddr *)&client_addr,&clilen);
	if(nfd==-1) {
		//hello ??
	}
	else {
		int a=0;
		for (a=0 ; a<16 ; a++)
			if ( clients[a] == -1 ) {
				clients[a]=nfd;
				break;
			}
		num_clients++;
		if ( debug )
			printf ("Added a lirc client (%i)\n", num_clients);
	}
}

void send_key(char *message)
{
	int i;
	for (i=0 ; i<16 ; i++) {
		if ( clients[i] != -1 ) {
			int done,todo,len;
			char *buf;

			len=strlen(message);
			buf=message;
			todo=len;

			if ( debug )
				printf ("  Sending to client. (%i)\n",i);

			while(todo) {
				done=write(clients[i],buf,todo);
				if(done<=0) {
					clients[i]=-1;
					num_clients--;
					break;
				}
				buf+=done;
				todo-=done;
			}
		}
	}
}

void fb_service() 
{
	int n=0;
	int retval;
	fd_set fds;
	fd_set efds;
	struct timeval tv;
	tv.tv_sec = 5;
	tv.tv_usec = 0;

	FD_ZERO(&fds);
	FD_ZERO(&efds);
	if (FB_present == 1) {
		FD_SET(rfd, &fds);
		FD_SET(rfd, &efds);
		n=rfd;
		tv.tv_sec = 30;
	}
	if (key_down == 1) {
		if ( debug )
			printf ("short select.\n");
		tv.tv_sec = 0;
		tv.tv_usec = 300000;
	}
	if(num_clients < 16) {
		FD_SET(sockfd, &fds);
		n = max(sockfd,n);
	}
	retval = select(n+1, &fds, NULL, &efds, &tv);
	if (retval == -1) {
		perror("select()");
	}
	else if (retval) {
		if( FD_ISSET(sockfd, &fds)) {
			if ( debug )
				printf ("Got a new client.\n");
			add_client();
		}
		if (FD_ISSET(rfd, &fds)) {
			int ret;
			if ( debug )
				printf ("Got a key \n");
			ret=get_key (rfd, lirc_message);

			if (ret==-1) {
				if ( debug )
					printf ("FB read failed. FB lost.\n");
				FB_present=0;
				close(rfd);
			}
			else if (ret==2){
				if ( debug )
					printf ("key unknown/multiple\n");
			}
			else if ( ret==1 ) {
				if ( debug )
					printf ("key down\n");
				key_down=1;
				if(num_clients > 0)
					send_key( lirc_message );
			}
			else if (ret==0){
				if ( debug )
					printf ("key up\n");
				key_down=0;
			}
		}
		if (FD_ISSET(rfd, &efds)) {
			if ( debug )
				printf ("FB lost.\n");
		}
		if ( debug )
			printf ("Select end.\n");
	}
	else if (!retval) {
		if ( debug )
			printf ("Select timeout.\n");
		if (key_down == 1) {
			if ( debug )
				printf ("auto-repeat.\n");
			send_key( lirc_message );
		}
	}
}

int main(int argc, char *argv[])
{
        int i ;
        int opt;


	while ((opt=getopt(argc,argv,"vdh"))!=-1){
                switch(opt){
                        case 'v':
			printf("%s\n",version);
			exit(0);
                        break;
                        case 'n':
			nofork=1;
                        break;
                        case 'd':
                        debug=1;
                        break;
                        case 'h':
                        help();
                        exit(0);
                        break;

        	default:
                        exit(0);
        	}
        }

        rfd=scanforfb("Sony Computer Entertainment Inc BD Remote Control");
        if (rfd!=0)
                FB_present=1;
        else
                FB_present=0;

        if (setupSocket() == -1) {
                printf("Can't init socket.\n");
                return -1;
        }

        /*Become a daemon process
        */
        if ( !debug  && !nofork ) {
                ipid=fork();
                if (ipid<0)
                        exit(1);
                if (ipid>0)
                        exit(0);

                setsid();

                i=open("/dev/null",O_RDWR);
                dup(i);
                dup(i);
        }
	else {
		if ( debug )
			printf ("Not forking to background.\n");
	}

        signal(SIGPIPE, SIG_IGN);

        clilen=sizeof(client_addr);


//	while (1) {
//		if ( debug )
//			printf ("Waiting for lirc client to connect.\n");

//		/*Wait until someone connects*/
//		fd=accept(sockfd,(struct sockaddr *)&client_addr,&clilen);
//		if(fd==-1) {
//			// Unfortunately this is deadly right now.
//			return -1;
//		}
//		else {
//			clients[0] = fd;
//			num_clients++;
//			nolinger(fd);
//		}

		/*This is the real duty cycle loop*/
		if ( debug )
			printf ("Starting main loop.\n");
		while(1)
		{
			while(FB_present == 0)
			{
				if ( debug )
					printf ("Looking for PS3 FB in busy loop.\n");
        			rfd=scanforfb("Sony Computer Entertainment Inc BD Remote Control");
        			if (rfd!=0) {
                			FB_present=1;
					break;
				}
			 	usleep(2000000);	
			}
			
			fb_service();
		}

//	}
}
